#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Author:
Benjamin Hills
University of Washington
Earth and Space Sciences
April 28, 2020
"""

import numpy as np
from constants import constants
from scipy.interpolate import interp1d

# Instantiate the constants class
const = constants()

# ------------------------------------------------------------------------------------------

def print_and_save(self,i,print_increment=1000):
    """
    Print and save output every 1000th time step
    """
    if i%print_increment== 0 or self.ts[i] == self.ts[-1]:
        if 'verbose' in self.flags:
            print('t =',int(self.ts[i]/const.spy),'; dt =',self.dt/const.spy,'; melt rate =',np.round(self.Mrate*1000.,2),'; melt cum = ',np.round(self.Mcum,2),'; q_b = ',self.q_b)
        if 'save_all' in self.flags:
            self.Mrate_all = np.append(self.Mrate_all,self.Mrate)
            self.Mcum_all = np.append(self.Mcum_all,self.Mcum)
            self.Ts_out = np.append(self.Ts_out,[self.T],axis=0)
            if self.Hs is not None:
                self.zs_out = np.append(self.zs_out,[self.z],axis=0)

# ------------------------------------------------------------------------------------------

def update_time(self,i):
    """
    Update variables to current time
    """
    self.Udef,self.Uslide = self.Udefs[i],self.Uslides[i]   # Update the velocity terms from input
    self.Ts,self.adot = self.Ts_s[i],self.adot_s[i]   # Update the velocity terms from input
    self.T[-1] = self.Ts                                 # Set surface temperature condition from input
    if self.Hs is not None:
        thickness_update(self,self.Hs[i]) # thickness update
    v_z_surf = self.adot      # set vertical velocity
    # add extra term from Weertman if desired
    if 'weertman_vel' in self.flags:
        v_z_surf += self.Udef*self.dH
    if self.p == 0.: # by exponent, gamma
        self.v_z = self.Mrate/const.spy + v_z_surf*(self.z/self.H)**self.gamma
    else: # by shape factor, p
        zeta = (1.-(self.z/self.H))
        self.v_z = self.Mrate/const.spy + v_z_surf*(1.-((self.p+2.)/(self.p+1.))*zeta+(1./(self.p+1.))*zeta**(self.p+2.))
    for i in range(len(self.z)):
        adv = (-self.v_z[i]*self.dt/self.dz)
        self.B[i,i] = adv
        self.B[i,i-1] = -adv
    # Ice Properties
    if 'temp-dependent' in self.flags:
        diffusivity_update(self)
    # Boundary Conditions
    self.B[0,:] = 0.  # Neumann at bed
    self.B[-1,:] = 0. # Dirichlet at surface
    # Source
    self.source_terms()
    self.Tgrad = -(self.qgeo+self.q_b)/self.k[0]             # Temperature gradient at bed
    self.Sdot[0] += -2*self.dz*self.Tgrad*self.diff[0]/self.dt # update boundaries on heat source vector
    self.Sdot[-1] = 0.

# ------------------------------------------------------------------------------------------

def melt_rate(self):
    """
    Calculate the melt/freeze rate and save the cumulative melt to an output field
    """
    ### Calculate the volume melted/frozen during the time step, then hard reset to pmp.
    if np.any(self.T>self.pmp): # If Melting
        Tplus = (self.T[self.T>self.pmp]-self.pmp[self.T>self.pmp])*self.int_stencil[self.T>self.pmp]*self.dz # Integrate the temperature above PMP (units- degCm)
        rho = self.rho[self.T>self.pmp]
        Cp = self.Cp[self.T>self.pmp]
        self.Mrate = np.sum(Tplus*rho*Cp*const.spy/(const.rhow*const.L*self.dt)) # calculate the melt rate in m/yr
        self.T[self.T>self.pmp] = self.pmp[self.T>self.pmp] # reset temp to PMP
        self.Mcum += self.Mrate*self.dt/const.spy # Update the cumulative melt by the melt rate
    elif self.Mcum > 0 and 'water_cum' in self.flags: # If freezing
        Tminus = (self.T[0]-self.pmp[0])*0.5*self.dz # temperature below the PMP; this is only for the point at the bed because we assume water drains
        rho = self.rho[0]
        Cp = self.Cp[0]
        self.Mrate = Tminus*rho*Cp*const.spy/(const.rhow*const.L*self.dt) # melt rate should be negative now.
        if self.Mrate*self.dt/const.spy < self.Mcum: # If the amount frozen this time step is less than water available
            self.T[0] = self.pmp[0] # reset to PMP
            self.Mcum += self.Mrate*self.dt/const.spy # Update the cumulative melt by the melt rate
        else: # If the amount frozen this time step is more than water available
            M_ = (self.Mrate*self.dt/const.spy-self.Mcum) # calculate the 'extra' amount frozen in m
            Tminus = M_*(const.rhow*const.L)/(self.rho*self.Cp) # What is the equivalent temperature to cool bottom node (units - degCm)
            self.T[0] = Tminus/(0.5*self.dz) + self.pmp[0] # update the temperature at the bed
            self.Mcum = 0. # cumulative melt to zero because everything is frozen
    else:
        self.Mrate = 0.
    # Cap the lake level
    if self.Mcum_max is not None:
        if self.Mcum > self.Mcum_max:
            self.Mcum = self.Mcum_max

def thickness_update(self,H_new,T_upper=None):
    """
    Stretch/shrink the depth array to match a new thickness at each timestep.
    Interpolate temperatures to their new position.
    """

    # If no upper fill value is provided for the interpolation, use the current surface temperature
    if T_upper is None:
        T_upper = self.T[-1]
    # Build an interpolator from the prior state
    T_interp = interp1d(self.z,self.T,fill_value=(np.nan,T_upper),bounds_error=False)
    # Interpolate for new temperatures
    self.z = np.linspace(0.,H_new,self.nz)
    self.T = T_interp(self.z)

    # Assign the new thickness value
    self.H = H_new

    # Update variables that are thickness dependent
    self.dz = np.mean(np.gradient(self.z))      # Vertical step
    self.P = const.rho*const.g*(self.H-self.z)  # Pressure
    self.pmp = self.P*self.beta                # Pressure melting

    # Stability, check the CFL
    if np.max(self.v_z)*self.dt/self.dz > 1.:
        print('CFL = ',max(self.v_z)*self.dt/self.dz,'; cannot be > 1.')
        print('dt = ',self.dt/const.spy,' years')
        print('dz = ',self.dz,' meters')
        raise ValueError("Numerically unstable, choose a smaller time step or a larger spatial step.")

    # Update diffusion stencil (advection gets updated with velocity profile)
    self.diff = (self.k/(self.rho*self.Cp))*(self.dt/(self.dz**2.))
    self.A.setdiag((1.-2.*self.diff)*np.ones(self.nz))       # Set the diagonal
    self.A.setdiag((1.*self.diff[1:])*np.ones(self.nz-1),k=-1)     # Set the diagonal
    self.A.setdiag((1.*self.diff[:-1])*np.ones(self.nz-1),k=1)      # Set the diagonal
    # Boundary Conditions
    # Neumann at bed
    self.A[0,1] = 2.*self.diff[0]
    # Dirichlet at surface
    self.A[-1,:] = 0.
    self.A[-1,-1] = 1.

# ------------------------------------------------------------------------------------------

def diffusivity_update(self):
    """
    Calculate the thermal diffusivity (k/rho/Cp) based on the updated temperature and density profile.
    Reset the stencils accordingly.
    """

    if 'conductivity' in self.flags:
        self.k = conductivity(self.T.copy(),self.rho)
    if 'heat_capacity' in self.flags:
        self.Cp = heat_capacity(self.T.copy(),self.rho)

    # Update diffusion stencil (advection gets updated with velocity profile)
    self.diff = (self.k/(self.rho*self.Cp))*(self.dt/(self.dz**2.))
    self.A.setdiag((1.-2.*self.diff)*np.ones(self.nz))       # Set the diagonal
    self.A.setdiag((1.*self.diff[1:])*np.ones(self.nz-1),k=-1)     # Set the diagonal
    self.A.setdiag((1.*self.diff[:-1])*np.ones(self.nz-1),k=1)      # Set the diagonal
    # Boundary Conditions
    # Neumann at bed
    self.A[0,1] = 2.*self.diff[0]
    # Dirichlet at surface
    self.A[-1,:] = 0.
    self.A[-1,-1] = 1.




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Author:
Benjamin Hills
University of Washington
Earth and Space Sciences
January 19, 2022
"""

import numpy as np
import matplotlib.pyplot as plt

from diffeq import rk4

# Import necessary libraries
import numpy as np
from scipy import sparse
from scipy.integrate import cumtrapz
from analytical_solutions import *
from ice_properties import *
from supporting_functions import *
from constants import constants

# Instantiate the constants class
const = constants()

class perol_ode():
    """
    A Runge Kutta solution to the 1-d steady-state heat equation with temperate ice.
    Perol and Rice (2015)
    """

    def __init__(self,Ts=-50.,H=2850.,adot=0.1,qgeo=0.050,p=1000.,
                 dS=0.,dTs=0.,dH=0.,da=0.,
                 eps_xy=0.,A_xy='full',
                 nz=101,tol=1e-5,const=const):
        """
        Initialize the model with constant terms

        Parameters
        ----------
        Ts:         float,  Surface Temperature (C)
        H:          float,  Ice thickness (m)
        adot:       float,  Accumulation rate (m/yr)
        qgeo:       float,  Geothermal Flux (W/m2)
        p:          float,  Lliboutry Shape Factor
        dS:         float,  Surface slope (degrees)
        eps_xy:     float,  Plane Strain Rate (yr-1)
        nz:         int,    Number of layers in the ice column
        const:      class,  Constants
        """



"""A variety of methods to solve ODE boundary value problems.

AUTHOR:
    Jonathan Senning <jonathan.senning@gordon.edu>
    Gordon College
    Based Octave functions written in the spring of 1999
    Python version: October-November 2008
"""

#-----------------------------------------------------------------------------

def shoot( f, a, b, z1, z2, t, tol ):
    """Implements the shooting method to solve second order BVPs

    USAGE:
        y = shoot(f, a, b, z1, z2, t, tol)

    INPUT:
        f     - function dy/dt = f(y,t).  Since we are solving a second-
                order boundary-value problem that has been transformed
                into a first order system, this function should return a
                1x2 array with the first entry equal to y and the second
                entry equal to y'.
        a     - solution value at the left boundary: a = y(t[0]).
        b     - solution value at the right boundary: b = y(t[n-1]).
        z1    - first initial estimate of y'(t[0]).
        z1    - second initial estimate of y'(t[0]).
        t     - array of n time values to determine y at.
        tol   - allowable tolerance on right boundary: | b - y[n-1] | < tol

    OUTPUT:
        y     - array of solution function values corresponding to the
                values in the supplied array t.

    NOTE:
        This function assumes that the second order BVP has been converted to
        a first order system of two equations.  The secant method is used to
        refine the initial values of y' used for the initial value problems.
    """


    max_iter = 25   # Maximum number of shooting iterations

    n = len( t )    # Determine the size of the arrays we will generate

    # Compute solution to first initial value problem (IVP) with y'(a) = z1.
    # Because we are using the secant method to refine our estimates of z =
    # y', we don't really need all the solution of the IVP, just the last
    # point of it -- this is saved in w1.

    y = rk4( f, [a,z1], t )
    w1 = y[n-1,0]

    print "%2d: z = %10.3e, error = %10.3e" % ( 0, z1, b - w1 )

    # Begin the main loop.  We will compute the solution of a second IVP and
    # then use the both solutions to refine our estimate of y'(a).  This
    # second solution then replaces the first and a new "second" solution is
    # generated.  This process continues until we either solve the problem to
    # within the specified tolerance or we exceed the maximum number of
    # allowable iterations.

    for i in xrange( max_iter ):

        # Solve second initial value problem, using y'(a) = z2.  We need to
        # retain the entire solution vector y since if y(t(n)) is close enough
        # to b for us to stop then the first column of y becomes our solution
        # vector.

        y = rk4( f, [a,z2], t )
        w2 = y[n-1,0]

        print "%2d: z = %10.3e, error = %10.3e" % ( i+1, z2, b - w2 )

        # Check to see if we are done...

        if abs( b - w2 ) < tol:
            break

        # Compute the new approximations to the initial value of the first
        # derivative.  We compute z2 using a linear fit through (z1,w1) and
        # (z2,w2) where w1 and w2 are the estimates at t=b of the initial
        # value problems solved above with y1'(a) = z1 and y2'(a) = z2.  The
        # new value for z1 is the old value of z2.

        #z1, z2 = ( z2, z1 + ( z2 - z1 ) / ( w2 - w1 ) * ( b - w1 ) )
        z1, z2 = ( z2, z2 + ( z2 - z1 ) / ( w2 - w1 ) * ( b - w2 ) )
        w1 = w2

    # All done.  Check to see if we really solved the problem, and then return
    # the solution.

    if abs( b - w2 ) >= tol:
        print "\a**** ERROR ****"
        print "Maximum number of iterations (%d) exceeded" % max_iter
        print "Returned values may not have desired accuracy"
        print "Error estimate of returned solution is %e" % ( b - w2 )

    return y[:,0]

#-----------------------------------------------------------------------------

def fd( u, v, w, t, a, b ):
    """Implements the shooting method to solve linear second order BVPs

    Compute finite difference solution to the BVP

        x'' = u(t) + v(t) x + w(t) x'
        x(t[0]) = a, x(t[n-1]) = b

    t should be passed in as an n element array.   u, v, and w should be
    either n element arrays corresponding to u(t), v(t) and w(t) or
    scalars, in which case an n element array with the given value is
    generated for each of them.

    USAGE:
        x = fd(u, v, w, t, a, b)

    INPUT:
        u,v,w - arrays containing u(t), v(t), and w(t) values.  May be
                specified as Python lists, NumPy arrays, or scalars.  In
                each case they are converted to NumPy arrays.
        t     - array of n time values to determine x at
        a     - solution value at the left boundary: a = x(t[0])
        b     - solution value at the right boundary: b = x(t[n-1])

    OUTPUT:
        x     - array of solution function values corresponding to the
                values in the supplied array t.
    """

    # Get the dimension of t and make sure that t is an n-element vector

    if type( t ) != numpy.ndarray:
        if type( t ) == list:
            t = numpy.array( t )
        else:
            t = numpy.array( [ float( t ) ] )

    n = len( t )

    # Make sure that u, v, and w are either scalars or n-element vectors.
    # If they are scalars then we create vectors with the scalar value in
    # each position.

    if type( u ) == int or type( u ) == float:
        u = numpy.array( [ float( u ) ] * n )

    if type( v ) == int or type( v ) == float:
        v = numpy.array( [ float( v ) ] * n )

    if type( w ) == int or type( w ) == float:
        w = numpy.array( [ float( w ) ] * n )

    # Compute the stepsize.  It is assumed that all elements in t are
    # equally spaced.

    h = t[1] - t[0];

    # Construct tridiagonal system; boundary conditions appear as first and
    # last equations in system.

    A = -( 1.0 + w[1:n] * h / 2.0 )
    A[-1] = 0.0

    C = -( 1.0 - w[0:n-1] * h / 2.0 )
    C[0] = 0.0

    D = 2.0 + h * h * v
    D[0] = D[n-1] = 1.0

    B = - h * h * u
    B[0] = a
    B[n-1] = b

    # Solve tridiagonal system

    for i in xrange( 1, n ):
        xmult = A[i-1] / D[i-1]
        D[i] = D[i] - xmult * C[i-1]
        B[i] = B[i] - xmult * B[i-1]

    x = numpy.zeros( n )
    x[n-1] = B[n-1] / D[n-1]

    for i in xrange( n - 2, -1, -1 ):
        x[i] = ( B[i] - C[i] * x[i+1] ) / D[i]

    return x

#-----------------------------------------------------------------------------

if __name__ == "__main__":

    import math
    from pylab import *

    # Solves x'' = x + 4exp(t), x(0)=1, x(1/2) = 2exp(1/2) using both the
    # finite difference method and the shooting method.

    # Set up interval.  We will solve the problem for both n=64 and n=128.

    a = 0.0
    b = 0.5
    n1 = 64
    n2 = 128
    t1 = linspace( a, b, n1 )
    t2 = linspace( a, b, n2 )

    # Compute exact solutions.  The transpose of the solution is taken because
    # both the finite difference function fd() and the shooting method function
    # shoot() return nx1 vectors rather than 1xn vectors.

    def exact(t):
        return exp(t) * ( 1 + 2 * t )

    x1 = exact( t1 )
    x2 = exact( t2 )

    # Compute finite difference solutions

    xfd1 = fd( 4 * exp( t1 ), 1, 0, t1, 1, 2 * exp( 0.5 ) )
    xfd2 = fd( 4 * exp( t2 ), 1, 0, t2, 1, 2 * exp( 0.5 ) )

    # Compute shooting method solutions

    def f(x,t):
        return array( [x[1], x[0]+4*exp(t)] )

    xs1 = shoot( f, exp(a), 2*exp(b), 3.0, 4.0, t1, 1e-5 )
    xs2 = shoot( f, exp(a), 2*exp(b), 3.0, 4.0, t2, 1e-5 )




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2021 bhills <benjaminhhills@gmail.com>
# Distributed under terms of the GNU GPL3.0 license.
"""
Skeleton import iceotherm/lib
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Author:
Benjamin Hills
University of Washington
Earth and Space Sciences
April 28, 2020
"""

import numpy as np
from scipy.optimize import minimize
from constants import constants

# ---------------------------------------------------

def conductivity(T,rho,const=constants()):
    """
    Temperature- and density-dependent conductivity

    Parameters
    --------
    T:      array
        Ice Temperature (C)
    rho:    array
        Firn density (kg m-3)
    const:  class,  Constants

    Output
    krho:   array, conductivity
    --------
    """
    if np.any(T<150):
        T += const.T0
    ki = 9.828*np.exp(-5.7e-3*T)
    krho = 2.*ki*rho/(3.*const.rho-rho)
    return krho

# ---------------------------------------------------

def heat_capacity(T,const=constants()):
    """
    Temperature--dependent heat capacity

    Parameters
    --------
    T:      array
        Ice Temperature (C)
    const:  class,  Constants

    Output
    --------
    Cp:     array, heat capacity
    """
    if np.any(T<150):
        T += const.T0
    Cp = 152.5 + 7.122*T
    return Cp

# ---------------------------------------------------

def rate_factor(T,const=constants(),
             z=None,H=None,pmp=0.,
             tau_xz=None,v_surf=None):
    """
    Rate Facor function for ice viscosity, A(T)
    Cuffey and Paterson (2010), equation 3.35

    Optional case for optimization to the surface velocity using function
    surf_vel_opt()

    Parameters
    ----------
    T:      array
        Ice Temperature (C)
    const:  class, optional
        Constants
    z:      array, optional
        Height (m)
    H:      float, optional
        Ice Thickness (m)
    pmp:      float, optional
        Pressure-Melting Point (C)
    tau_xz: array, optional
        Shear stress profile, only needed if optimizing the strain rate to match surface
    v_surf: float, optional
        Surface velocity to be matched in optimization

    Output
    ----------
    A:      array,  Rate Factor, viscosity = A^(-1/n)/2
    """

    # create an array for activation energies
    Qact = const.Qminus*np.ones_like(T)
    if hasattr(T,'__len__'):
        Qact[T>-10.] = const.Qplus
    elif T>-10.:
        Qact = const.Qplus
    # Overburden pressure
    if z is not None:
        pmp = const.rho*const.g*(H-z)*const.beta

    if v_surf is None:
        # rate factor Cuffey and Paterson (2010) equation 3.35
        A = const.Astar*np.exp(-(Qact/const.R)*((1./(T+const.T0-pmp))-(1./(const.Tstar-pmp))))
    else:
        # Get the final coefficient value
        res = minimize(surf_vel_opt, 1, args=(T,z,pmp,Qact,tau_xz,v_surf))
        # C was scaled for appropriate stepping of the minimization function, scale back
        C_fin = res['x']*const.Astar
        # rate factor Cuffey and Paterson (2010) equation 3.35
        A = C_fin*np.exp(-(Qact/const.R)*((1./(T+const.T0-pmp))-(1./(const.Tstar-pmp))))
    return A

# ---------------------------------------------------

def surf_vel_opt(C,T,z,pmp,Qact,tau_xz,v_surf,
                 const=constants()):
    """
    Optimize the viscosity profile using the known surface velocity

    Parameters
    ----------
    C:      float
        Rate factor coefficient (multiplier)
    T:      array
        Ice Temperature (C)
    z:      array
        Height (m)
    pmp:      float, optional
        Pressure-Melting Point (C)
    Qact:   float
        Activation Energy
    tau_xz: array
        Shear stress profile, only needed if optimizing the strain rate to match surface
    v_surf: float
        Surface velocity to be matched in optimization
    const:  class, optional
        Constants

    Output
    ----------
    energy_optimizer    float
        cost to be minimized
    """
    # Change the coefficient so that the minimization function takes appropriate steps
    C_opt = C*const.Astar
    # rate factor Cuffey and Paterson (2010) equation 3.35
    A = C_opt*np.exp(-(Qact/const.R)*((1./(T+const.T0-pmp))-(1/(const.Tstar-pmp))))
    # Shear Strain Rate, Weertman (1968) eq. 7
    eps_xz = A*tau_xz**const.n
    Q = 2.*(eps_xz*tau_xz)
    # Integrate the heat source to get total energy input
    Q_opt = np.trapz(Q,z)
    # Optimize to conserve energy
    energy_optimizer = abs(Q_opt-v_surf*tau_xz[0])/const.Astar
    return energy_optimizer



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Author:
Benjamin Hills
University of Washington
Earth and Space Sciences
April 28, 2020
"""

import numpy as np

class constants(object):
    """
    Temperature Constants

    Cuffey and Paterson (2010)
    """
    def __init__(self):
        # general
        self.spy  = 60.*60.*24.*365.24          # sec yr-1
        self.spd  = 60.*60.*24.                 # sec day-1
        self.g = 9.81                           # Gravity m s-2
        self.T0 = 273.15                        # Reference Tempearature, triple point for water, K
        self.R = 8.321                          # Gas Constant J mol-1 K-1
        self.kBoltz = 1.38064852e-23            # m2 kg s-2 K-1
        self.rhow = 1000.                       # Density of water kg m-3
        # CP (2010) pg. 72
        self.n = 3.                             # Creep Exponent
        self.Tstar = 263.                       # Reference Temperature K
        self.Qminus = 6.0e4                     # Activation Energy <10C J mol-1
        self.Qplus = 11.5e4                     # Activation Energy >10C J mol-1
        self.Astar = 3.5e-25                    # Creep Parameter Pa-3 s-1
        # CP (2010) pg. 12
        self.rho = 917.                         # Ice Density kg m-3
        # CP (2010) pg. 400
        self.Cp = 2097.                         # Specific Heat Capacity J kg-1 K-1
        self.L = 3.335e5                        # Latent Heat of Fusion J kg-1
        self.k = 2.1                            # Thermal Conductivity J m-1 K-1 s-1
        self.K = 1.09e-6                        # Thermal Diffusivity m2 s-1
        # CP (2010) pg. 406
        self.beta = -7.42e-8                    # Clausius-Clapeyron K Pa-1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Author:
Benjamin Hills
University of Washington
Earth and Space Sciences
April 28, 2020
"""

# Import necessary libraries
import numpy as np
from scipy import sparse
from scipy.integrate import cumtrapz
from analytical_solutions import *
from ice_properties import *
from supporting_functions import *
from constants import constants

# Instantiate the constants class
const = constants()

class ice_temperature():
    """
    1-D finite-difference model for ice temperature based on
    Weertman (1968)

    Features:
        - Initialization
            1-D Analytical temperature profile from either Robin (1955) or Rezvanbehbahani et al. (2019)
            then spin up the numbercal model to steady state until all points are changing by less then 'self.tol' every time step.
        - Vertical Advection
            Two options:
                1) Use an exponential form as in Rezvanbehbahani et al. (2019)
                2) Use the shape factor, p, as from Lliboutry
        - Horizontal Advection
            Longitudinal temperature, accumulation, and thickness gradients must be input manually.
            Shear stress for lamellar flow then optimize the rate_factor to match the surface velocity.
        - Strain Heating
            Option for additional xy plane shear component (i.e. Meyer/Perol models)
        - Melt
    """

    def __init__(self,Ts=-50.,H=2850.,adot=0.1,qgeo=0.050,p=1000.,
                 dS=0.,dTs=0.,dH=0.,da=0.,
                 eps_xy=0.,A_xy='full',
                 nz=101,tol=1e-5,const=const):
        """
        Initialize the model with constant terms

        Parameters
        ----------
        Ts:         float,  Surface Temperature (C)
        H:          float,  Ice thickness (m)
        adot:       float,  Accumulation rate (m/yr)
        qgeo:       float,  Geothermal Flux (W/m2)
        p:          float,  Lliboutry Shape Factor
        dS:         float,  Surface slope (degrees)
        eps_xy:     float,  Plane Strain Rate (yr-1)
        nz:         int,    Number of layers in the ice column
        const:      class,  Constants
        """

        ### Numerical Inputs ###
        self.tol = tol                  # Convergence criteria
        self.nz=nz                      # Number of layers in the ice column

        ### Boundary Constraints ###
        self.Ts = Ts                    # Surface Temperature   [C]
        self.qgeo = qgeo                # Geothermal flux       [W/m2]
        self.H = H                      # Ice thickness         [m]
        self.adot = adot/const.spy      # Accumulation rate     [m/s]
        self.gamma = None               # Exponent for vertical velocity; if None calculate using Rezvanbehbahani (2019) eq. 19
        self.p = p                      # Lliboutry shape factor for vertical velocity (large p is ~linear)

        ### Internal Constraints ###
        self.eps_xy = eps_xy/const.spy  # Plane Strain rate [s-1]
        self.A_xy_init = A_xy           # Rate factor for plane strain heat source

        ### Ice Properties ###
        self.beta = const.beta                  # Melting point depression (default to const.beta)  [K/Pa]
        self.k = const.k*np.ones(self.nz)       # Thermal conductivity (default to const.k)         [W/m/K]
        self.Cp = const.Cp*np.ones(self.nz)     # Heat capacity (default to const.Cp)               [J/kg/K]
        self.rho = const.rho*np.ones(self.nz)   # Density (default to const.rho)                    [kg/m3]

        ### Gradients ###
        self.dS = np.sin(dS*np.pi/180.) # Surface gradient in x/y directions, used for deformational flow calculation [m/m]
        self.dTs = dTs                  # Change in air temperature over distance x/y   [C/m]
        self.dH = dH                    # Thickness gradient in x/y directions          [m/m]
        self.da = da/const.spy          # Accumulation gradient in x/y directions       [m/yr/m]

        ### Velocity Terms ###
        self.Udef = 0.                  # Deformational velocity    [m/s]
        self.Uslide = 0.                # Sliding velocity          [m/s]

        ### Thickness over time (default to None) ###
        self.Hs = None                  # Array of ice thicknesses  [m]

        ### Melting Conditions ###
        self.Mrate = 0.                 # Melt rate                             [m/s]
        self.Mcum = 0.                  # Cumulative melt                       [m]
        self.Mcum_max = None            # Max Cumulative melt for a capped lake [m]

        ### Discretize the vertical coordinate ###
        self.z = np.linspace(0,self.H,self.nz)
        self.dz = np.mean(np.gradient(self.z))      # Vertical step
        self.P = const.rho*const.g*(self.H-self.z)  # Pressure
        self.pmp = self.P*self.beta                # Pressure melting

        ### Empty Time Array as Default ###
        self.ts=[]

        ### Flags ###
        self.flags = ['verbose']

    # ------------------------------------------------------------------------------------------

    def initial_conditions(self,const=const,analytical=Robin_T):
        """
        Define the initial ice column properties using an analytical solution
        with paramaters from the beginning of the time series.
        """

        # initial temperature from analytical solution
        if analytical==Robin_T:
            self.T,M = analytical(self)
        else:
            self.T = analytical(self)

        # vertical velocity
        if self.p is None:
            # by exponent, gamma
            self.v_z = self.adot*(self.z/self.H)**self.gamma
        else:
            # by shape factor, p
            zeta = (1.-(self.z/self.H))
            self.v_z = self.adot*(1.-((self.p+2.)/(self.p+1.))*zeta+(1./(self.p+1.))*zeta**(self.p+2.))


    def source_terms(self,const=const):
        """
        Heat sources from strain heating and downstream advection (this is typically a heat sink)
        """

        # Shear Stress by Lamellar Flow (van der Veen section 4.2)
        tau_xz = const.rho*const.g*(self.H-self.z)*abs(self.dS)     # [Pa]
        # Sliding friction heat production
        self.tau_b = tau_xz[0]                          # [Pa]
        self.q_b = self.tau_b*self.Uslide               # [W/m2]

        # Initialize source term at zeros
        eps_xz = np.zeros_like(tau_xz)
        self.Sdot = np.zeros_like(tau_xz)

        ### Vertical Shear Heat Production ###
        if 'vertical_shear' in self.flags:
            # Calculate the rate_factor
            A = rate_factor(self.T,z=self.z,H=self.H,const=const,tau_xz=tau_xz,v_surf=self.Udef)   # [/s/Pa3]
            # Strain rate, Weertman (1968) eq. 7
            eps_xz = (A*tau_xz**const.n)                # [/s]
            # strain heat term
            Q = 2.*(eps_xz*tau_xz)/(self.rho*self.Cp)   # [K/s]
            self.Sdot += Q

        ### Plane Strain Heat Production ###
        if 'plane_strain' in self.flags:
            # Calculate the rate_factor
            if self.A_xy_init == 'full':
                self.A_xy = rate_factor(self.T,z=self.z,H=self.H,const=const)
            elif self.A_xy_init == 'temperate':
                self.A_xy = rate_factor(self.pmp,z=self.z,H=self.H,const=const)
            else:
                self.A_xy = self.A_xy_init
            tau_xy = (self.eps_xy/self.A_xy)**(1./const.n)
            Q_xy = 2.*(self.eps_xy*tau_xy)/(self.rho*self.Cp)
            # Add to the source term
            self.Sdot += Q_xy

        ### Advective Heat Sink ###
        if 'long_advection' in self.flags:
            v_x = self.Uslide + np.insert(cumtrapz(eps_xz,self.z),0,0)    # Horizontal velocity
            # Horizontal Temperature Gradients, Weertman (1968) eq. 6b
            dTdx = self.dTs + (self.T-self.Ts)/2. * (self.dH/self.H - self.da/self.adot)
            # Final Source Term
            self.Sdot -= v_x*dTdx

    # ------------------------------------------------------------------------------------------

    def stencil(self,dt=None,const=const):
        """
        Finite Difference Scheme for 1-d advection diffusion
        Surface boundary is fixed (air temperature)
        Bed boundary is gradient (geothermal flux plus sliding heat source)
        """

        # Choose time step
        if dt is None:
            # Check if the time series is monotonically increasing
            if len(self.ts) == 0:
                raise ValueError("If no time array, must input a time step or use 'CFL'.")
            if not np.all(np.gradient(np.gradient(self.ts))<self.tol):
                raise ValueError("Time series must monotonically increase.")
            self.dt = np.mean(np.gradient(self.ts))
        elif dt == 'CFL':
            # set time step with CFL
            self.dt = 0.5*self.dz/np.max(self.v_z)
        else:
            self.dt = dt
        # Stability, check the CFL
        if np.max(self.v_z)*self.dt/self.dz > 1.:
            print('CFL = ',max(self.v_z)*self.dt/self.dz,'; cannot be > 1.')
            print('dt = ',self.dt/const.spy,' years')
            print('dz = ',self.dz,' meters')
            raise ValueError("Numerically unstable, choose a smaller time step or a larger spatial step.")


        ### Stencils ###

        # Diffusion Matrix
        self.diff = (self.k/(self.rho*self.Cp))*(self.dt/(self.dz**2.)) # Thermal diffusivity (adjusted for time and spatial step)
        self.A = sparse.lil_matrix((self.nz, self.nz))                  # Create a sparse matrix
        self.A.setdiag((1.-2.*self.diff)*np.ones(self.nz))              # Set the center diagonal (centered difference)
        self.A.setdiag((1.*self.diff[1:])*np.ones(self.nz-1),k=-1)      # Set the -1 diagonal
        self.A.setdiag((1.*self.diff[:-1])*np.ones(self.nz-1),k=1)      # Set the +1 diagonal
        self.B = sparse.lil_matrix((self.nz, self.nz))                  # Create a sparse matrix

        # Advection Matrix
        for i in range(len(self.z)):
            adv = (-self.v_z[i]*self.dt/self.dz)    # Advection (adjusted for time and spatial step)
            self.B[i,i] = adv                       # Set the center diagonal
            self.B[i,i-1] = -adv                    # Set the -1 diagonal (upwind scheme)

        # Boundary Conditions
        # Neumann at bed
        self.A[0,1] = 2.*self.diff[0]
        self.B[0,:] = 0.
        # Dirichlet at surface
        self.A[-1,:] = 0.
        self.A[-1,-1] = 1.
        self.B[-1,:] = 0.

        # Source Term
        if 'Sdot' not in vars(self):
            raise ValueError('Must run the source_terms function before defining the stencil.')
        self.Tgrad = -(self.qgeo+self.q_b)/self.k[0]                    # Temperature gradient at bed
        self.Sdot[0] += -2.*self.dz*self.Tgrad*self.diff[0]/self.dt     # Set the source term at the bed to match the appropriate Temperature gradient
        self.Sdot[-1] = 0.                                              # Set the source term at the surface to 0.

        # Integration stencil to calculate melt volume near the bottom of the profile
        self.int_stencil = np.ones_like(self.z)
        self.int_stencil[[0,-1]] = 0.5              # Lowest point only represents half of a full spatial step (dz/2)

    # ------------------------------------------------------------------------------------------

    def numerical_to_steady_state(self,const=const):
        """
        Run the initial conditions until stable within self.tol
        """

        if 'Sdot' not in vars(self):
            raise ValueError('Must calculate the source_terms before running to steady state.')
        if 'A' not in vars(self):
            raise ValueError('Must calculate the stencil before running to steady state.')

        # Calculate the first updated temperature profile using the stencils and heat sources calculated above.
        T_new = self.A*self.T - self.B*self.T + self.dt*self.Sdot
        # Iteration counter
        steady_iter = 0
        if 'verbose' in self.flags:
            print('Running model to steady state')
        # Continue to iterate until the updated temperature profile is all within self.tol
        while steady_iter < 1000 or any(abs(self.T[1:]-T_new[1:])>self.tol):
            # Print output
            if 'verbose' in self.flags and steady_iter%1000==0:
                print('.',end='')
            self.T = T_new.copy()
            # Update the thermal diffusivity based on the new temperature profile
            if 'temp-dependent' in self.flags:
                diffusivity_update(self)
            if 'long_advection' in self.flags and steady_iter%1000==0:
                self.source_terms()
                self.stencil(self.dt)
            # Calculate the updated temperature profile using the stencils and heat sources
            T_new = self.A*self.T - self.B*self.T + self.dt*self.Sdot
            # Reset anything above the pressure melting point
            T_new[T_new>self.pmp] = self.pmp[T_new>self.pmp]
            # Update the iteration counter
            steady_iter += 1
        self.T = T_new.copy()
        # Print one more line to break the '...'
        if 'verbose' in self.flags:
            print('')

        # Run one final iteration to see how much things are changing still
        self.T = self.A*self.T - self.B*self.T + self.dt*self.Sdot
        melt_rate(self)
        self.T_steady = self.T.copy()

    # ------------------------------------------------------------------------------------------

    def numerical_transient(self,const=const,*args,**kwargs):
        """
        Non-Steady Model
        Run the finite-difference model as it has been set up through the other functions.
        """

        # Set up the output arrays
        if 'save_all' in self.flags:
            # Output temperature profiles
            self.Ts_out = np.empty((0,len(self.T)))
            # Output melt rates
            self.Mrate_all = np.empty((0))
            # Output cumulative melt
            self.Mcum_all = np.array([0])
            # Output the depth arrays
            if self.Hs is not None:
                self.zs_out = np.empty((0,len(self.z)))

        # Expand the deformational and sliding velocity terms into an array if that has not been added manually yet
        if len(self.ts)>0 and 'Udefs' not in vars(self):
            if 'verbose' in self.flags:
                print('No velocity arrays set, setting to constant value.')
            self.Udefs, self.Uslides = self.Udef*np.ones_like(self.ts), self.Uslide*np.ones_like(self.ts)

        # Iterate through all times
        for i in range(len(self.ts)):

            ### Print and output
            print_and_save(self,i,*args,**kwargs)

            ### Update to current time
            update_time(self,i)

            ### Solve (forward difference in time)
            T_new = self.A*self.T - self.B*self.T + self.dt*self.Sdot
            self.T = T_new

            ### Calculate melting/freezing
            melt_rate(self)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2021 bhills <benjaminhhills@gmail.com>
# Distributed under terms of the GNU GPL3.0 license.

from setuptools import setup

version = '1.0.1'

packages = ['iceotherm',
            'iceotherm.lib']

requires = ['numpy',
            'matplotlib',
            'scipy']

setup(
    name='iceotherm',
    version=version,
    description='A Python package for glacier and ice-sheet thermal modeling',
    url='https://github.com/benhills/iceotherm',
    author='Benjamin Hills',
    author_email='bhills@uw.edu',
    license='GNU GPL-3.0',
    packages=packages,
    install_requires=requires
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Author:
Benjamin Hills
University of Washington
Earth and Space Sciences
Jan 5, 2022
"""

import numpy as np
from scipy.special import gamma as γ
from scipy.special import gammaincc as γincc
from scipy.special import erf
from scipy.integrate import quad
from scipy.special import lambertw
from constants import constants

from ice_properties import *

def Robin_T(m,T_bulk=None,const=constants(),melt=True,verbose=False):
    """
    Analytic ice temperature model from Robin (1955)

    Assumptions:
        1) no horizontal advection
        2) vertical advection is linear with depth
        3) firn column is treated as equivalent thickness of ice
        4) If base is warmer than the melting temperature recalculate with new basal gradient
        5) no strain heating

    Parameters
    ----------
    m:          class, Model
    T_bulk      float, Temperature input to the rate factor function, A(T)
    const:      class,  Constants
    melt:       bool,   Choice to allow melting, when true the bed temperature
                        is locked at the pressure melting point and melt rates
                        are calculated
    verbose:    bool, option to print all output

    Output
    ----------
    T:          1-D array,  Analytic solution for ice temperature
    M:          float, Melt rate
    """

    # Thermal constants
    if T_bulk is None:
        k = const.k
        Cp = const.Cp
    else:
        if T_bulk == 'average':
            T_bulk = np.mean([m.Ts,m.pmp[0]])
        k = conductivity(T_bulk,const.rho)
        Cp = heat_capacity(T_bulk)
    alpha = k/(const.rho*Cp)

    q2 = m.adot/(2*alpha*m.H)
    Tb_grad = -m.qgeo/k
    f = lambda z : np.exp(-(z**2.)*q2)
    TTb = Tb_grad*np.array([quad(f,0,zi)[0] for zi in m.z])
    dTs = m.Ts - TTb[-1]
    T = TTb + dTs
    # recalculate if basal temperature is above melting (see van der Veen pg 148)
    if melt and T[0] > m.pmp[0]:
        Tb_grad = -2.*np.sqrt(q2)*(m.pmp[0]-m.Ts)/np.sqrt(np.pi)*(np.sqrt(erf(m.adot*m.H/(2.*alpha)))**(-1))
        TTb = Tb_grad*np.array([quad(f,0,zi)[0] for zi in m.z])
        dTs = m.Ts - TTb[-1]
        T = TTb + dTs
        M = (Tb_grad + m.qgeo/k)*k/const.L
        if verbose:
            print('Melting at the bed: ', np.round(M*const.spy/const.rho*1000.,2), 'mm/year')
    else:
        M = 0.
    if verbose:
        print('Finished Robin Solution for analytic temperature profile.\n')
    return T,M

# ---------------------------------------------------

def Rezvan_T(m,const=constants(),rate_factor=rate_factor,T_bulk=-10.,tau_dx=0.,verbose=False):
    """
    1-D Analytical temperature profile from Rezvanbehbahani et al. (2019)
    Main improvement from the Robin (1955) solution is the nonlinear vertical velocity profile

    Assumptions:
        1) no horizontal advection
        2) vertical advection takes the form v=(z/H)**(gamma)
        3) firn column is treated as equivalent thickness of ice
        TODO: 4) If base is warmer than the melting temperature recalculate with new basal gradient
        5) strain heating is added to the geothermal flux

    Parameters
    ----------
    m:              class,      Model
    const:          class,      Constants
    rate_factor:    function,   Calculate the rate factor from Glen's Flow Law
    T_bulk:         float,      Temperature input to rate factor function (C)
    tau_dx:         float,      Driving stress input directly (Pa)
    gamma_plus:     bool,       Optional, Determine gama_plus from the logarithmic regression with Pe Number
    verbose:        bool,       Print all output

    Output
    ----------
    T:              1-D array,  Analytic solution for ice temperature
    """

    # Thermal constants
    if T_bulk is None:
        k = const.k
        Cp = const.Cp
        A = const.Astar
    else:
        if T_bulk == 'average':
            T_bulk = np.mean([m.Ts,m.pmp[0]])
        k = conductivity(T_bulk,const.rho)
        Cp = heat_capacity(T_bulk)
        A = rate_factor(T_bulk,const)
    alpha = k/(const.rho*Cp)

    if m.gamma is None:
        # Solve for gamma using the logarithmic regression with the Pe number
        Pe = m.adot*m.H/alpha
        if Pe < 5. and verbose:
            print('Pe:',Pe)
            print('The gamma_plus fit is not well-adjusted for low Pe numbers.')
        # Rezvanbehbahani (2019) eq. (19)
        m.gamma = 1.39+.044*np.log(Pe)
    if tau_dx == 0.:
        # driving stress Nye (1952)
        tau_dx = const.rho*const.g*m.H*np.sin(m.dS)
    # Rezvanbehbahani (2019) eq. (22)
    qgeo_s = (2./5.)*A*m.H*tau_dx**4.
    qgeo = m.qgeo + qgeo_s
    # Rezvanbehbahani (2019) eq. (19)
    lamb = m.adot/(alpha*m.H**m.gamma)
    phi = -lamb/(m.gamma+1)

    # Rezvanbehbahani (2019) eq. (17)
    Γ_1 = γincc(1/(1+m.gamma),-phi*m.z**(m.gamma+1))*γ(1/(1+m.gamma))
    Γ_2 = γincc(1/(1+m.gamma),-phi*m.H**(m.gamma+1))*γ(1/(1+m.gamma))
    term2 = Γ_1-Γ_2
    T = m.Ts + m.qgeo*(-phi)**(-1./(m.gamma+1.))/(k*(m.gamma+1))*term2
    return T

# ---------------------------------------------------

def Meyer_T(m,const=constants(),
            rate_factor=rate_factor,
            T_bulk='average',
            Tb=0.,lam=0.,
            verbose=False):
    """
    Meyer and Minchew (2018)
    A 1-D analytical model of temperate ice in shear margins
    Uses the contact problem in applied mathematics

    Assumptions:
        1) Horizontal advection is treated as an energy sink by subtracting from the shear heat source
        2) Vertical advection is constant in depth (they do some linear analysis in their supplement)
        3) All ice properties, including rate factor A, are constant
        4) Ice base is at the melting temperature
        5) Melting temperature is 0 through the column

    Parameters
    ----------
    m:              class,  Model
    const:          class,  Constants
    rateFactor:     func,   function for the rate factor, A in Glen's Law
    T_bulk:         float, Temperature input to rate factor function (C)
    Tb:             float,  Basal temperature, at the pressure melting point
    lam:            float,  Paramaterized horizontal advection term
                            Meyer and Minchew (2018) eq. 11
    verbose:        bool, option to print all output

    Output
    ----------
    T:              1-D array,  Analytic solution for ice temperature
    """

    # Thermal constants
    if T_bulk is None:
        k = const.k
        Cp = const.Cp
        # rate factor (Meyer uses 2.4e-24; Table 1)
        A = 2.4e-24
    else:
        if T_bulk == 'average':
            T_bulk = np.mean([m.Ts,m.pmp[0]])
        k = conductivity(T_bulk,const.rho)
        Cp = heat_capacity(T_bulk)
        A = rate_factor(T_bulk,const)
    alpha = k/(const.rho*Cp)

    # Brinkman Number
    S = 2.*A**(-1./const.n)*(m.eps_xy)**((const.n+1.)/const.n)
    dT = Tb - m.Ts
    Br = (S*m.H**2.)/(k*dT)
    # Peclet Number
    Pe = (const.rho*Cp*m.adot*m.H)/(k)
    LAM = lam*m.H**2./(k*dT)
    if verbose:
        print('Meyer; Pe:', Pe,'Br:',Br)
    # temperature solution is different for diffusion only vs. advection-diffusion
    if abs(Pe) < 1e-3:
        # Critical Shear Strain
        eps_bar = (k*dT/(A**(-1/const.n)*m.H**(2.)))**(const.n/(const.n+1.))
        # Find the temperate thickness
        if m.eps_xy > eps_bar:
            hbar = 1.-np.sqrt(2./Br)
        else:
            hbar = 0.
        # Solve for the temperature profile
        T = m.Ts + dT*(Br/2.)*(1.-((m.z/m.H)**2.)-2.*hbar*(1.-m.z/m.H))
        T[m.z/m.H<hbar] = 0.
    else:
        # Critical Shear Strain
        eps_1 = (((0.5*Pe**2.)/(Pe-1.+np.exp(-Pe))+0.5*LAM)**(const.n/(const.n+1.)))
        eps_bar = eps_1 * ((k*dT/(A**(-1./const.n)*m.H**(2.)))**(const.n/(const.n+1.)))
        # Find the temperate thickness
        if m.eps_xy > eps_bar:
            h_1 = 1.-(Pe/(Br-LAM))
            h_2 = -(1./Pe)*(1.+np.real(lambertw(-np.exp(-(Pe**2./(Br-LAM))-1.))))
            hbar = h_1 + h_2
        else:
            hbar = 0.
        T = m.Ts + dT*((Br-LAM)/Pe)*(1.-m.z/m.H+(1./Pe)*np.exp(Pe*(hbar-1.))-(1./Pe)*np.exp(Pe*((hbar-m.z/m.H))))
        T[m.z/m.H<hbar] = 0.
    return T

# ---------------------------------------------------

def Perol_T(m,const=constants(),
                rate_factor=rate_factor,
                T_bulk='average',
                verbose=False):
    """
    Perol and Rice (2015)
    Analytic Solution for temperate ice in shear margins (equation #5)

    Assumptions:
        1) Bed is at the melting point
        2) All constants are temperature independent

    Parameters
    ----------
    m:          class,  Model
    const:      class,  Constants
    rateFactor: func,   function for the rate factor, A in Glen's Law
    T_bulk      float, Temperature input to the rate factor function, A(T)
    verbose:    bool, option to print all output

    Output
    ----------
    T:          1-D array,  Analytic solution for ice temperature
    """

    # Thermal constants
    if T_bulk is None:
        k = const.k
        Cp = const.Cp
        # rate factor (Meyer uses 2.4e-24; Table 1)
        A = 2.4e-24
    else:
        if T_bulk == 'average':
            T_bulk = np.mean([m.Ts,m.pmp[0]])
        k = conductivity(T_bulk,const.rho)
        Cp = heat_capacity(T_bulk)
        A = rate_factor(T_bulk,const)
    alpha = k/(const.rho*Cp)

    # Peclet Number
    Pe = m.adot*m.H/(k/(const.rho*Cp))
    # Strain Heating
    S = 2.*A**(-1./const.n)*(m.eps_xy)**((const.n+1.)/const.n)
    if verbose:
        print('Perol; A:',A, 'S:',S)
    # Empty Array for Temperatures, then loop through all z's
    T = np.empty_like(m.z)
    for i in range(len(m.z)):
        # Two functions to be integrated
        def f1(lam):
            return (1.-np.exp(-lam*Pe*m.z[i]**2./(2.*m.H**2.)))/(2.*lam*np.sqrt(1.-lam))
        def f2(lam):
            return (1.-np.exp(-lam*Pe/2.))/(2.*lam*np.sqrt(1.-lam))
        # Calculate temperature profile
        T[i] = m.pmp[0] + (m.Ts-m.pmp[0])*erf(np.sqrt(Pe/2.)*(m.z[i]/m.H))/erf(np.sqrt(Pe/2.)) - \
            S*m.H**2./(k*Pe) * (quad(f1,0.,1.)[0] - \
            (erf(np.sqrt(Pe/2.)*(m.z[i]/m.H))/erf(np.sqrt(Pe/2.))) * quad(f2,0.,1.)[0])
    return T
